#!/usr/bin/python

# HTMLProc
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Library General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
#
# (C)2004 Raster Software Vigo

import string
import sys
import dircache
import os
import shutil
import subprocess

class directory_walker:

	def __init__(self,path):
	
		self.directories=[]
		self.init_path=path
		self.directories.append(path)
	
	def take_directory_listing(self):
	
		if len(self.directories)==0:
			return self.directories # if there's no more files, return an empty list
		
		current_dir=self.directories[0]
		
		list=dircache.listdir(current_dir)
		list=list[:] # copy it to allow modification
		dircache.annotate(current_dir,list)		
		salida=[current_dir]
		for elemento in list:
			if elemento[0]==".":
				continue # don't add hidden files or directories
			if elemento[-1]=="/": # if an element is a directory, add it to the list
				self.directories.append(current_dir+elemento)
			else:
				salida.append(elemento) # if not, add it to the list of files				

		self.directories=self.directories[1:] # erase the first element because it's done
		return salida # return the files with the path as first element


class processor:

	extensions=[] #list of file extensions to process
	pictures=[] # list of file extensions for pictures
	tolink=[] # list of file extensions for other files (tar.bz, zip...)
	protocols=[] # list of internet protocols (ftp, http...)
	exclude=[] # list of file extensions to not process when included
	identify_pic=False # by default, don't identify pictures
	srcdir="" # directory with source files
	finaldir="" # directory where we will store the processed files
	diccionario={} # dictionary with pictures and its size
	ficheros={} # dictionary with filenames and its path
	path_source=""
	makecopy=False

	def __init__(self,configfile):

		if configfile=="":
			return # if the file is empty, we will configure all with the method
		
		if os.path.isfile(configfile)==False:
			print "Can't find the config file "+configfile
			sys.exit()
		
		fichero=open(configfile,"r")
		
		modo=0
		self.__class__.identify_pic=False
		
		for linea2 in fichero:
			if linea2[-1]=="\n":
				linea=linea2[:-1] #jump over the carriage return at end
			else:
				linea=linea2
			if linea=="\n" or linea=="" or linea[0]=="#":
				continue
			if linea[0]==":":
				if linea==":source":
					modo=1
				elif linea==":output":
					modo=2
				elif linea==":files":
					modo=3
				elif linea==":pictures":
					modo=4
				elif linea==":binaries":
					modo=5
				elif linea==":exclude":
					modo=6
				elif linea==":protocols":
					modo=7
				elif linea==":identify":
					self.__class__.identify_pic=True
					modo=0
				elif linea==":copyfiles":
					self.__class__.makecopy=True
					modo=0
				else:
					modo=0
			else:
				if modo==1:
					self.__class__.srcdir=linea
				elif modo==2:
					self.__class__.finaldir=linea
				elif modo==3:
					self.extensions.append(linea)
				elif modo==4:
					self.pictures.append(linea)
				elif modo==5:
					self.tolink.append(linea)
				elif modo==6:
					self.exclude.append(linea)
				elif modo==7:
					self.protocols.append(linea)
		fichero.close()

		if self.srcdir=="":
			print "The config file doesn't have a source directory. Fix it."
			sys.exit()
		
		if self.finaldir=="":
			print "The config file doesn't have an output directory. Fix it."
			sys.exit()
		
		if os.path.isdir(self.srcdir)==False:
			print "The source directory doesn't exists. Fix it."
			sys.exit()

		if os.path.isdir(self.finaldir)==False:
			print "The output directory doesn't exists. Fix it."
			sys.exit()
			
		if self.srcdir[-1]!="/":
			self.srcdir=self.srcdir+"/" # directories must end in /, so we add it if needed
			
		if self.finaldir[-1]!="/":
			self.finaldir=self.finaldir+"/"
		
		# prepare the file list with its directories
		self.path_source=self.srcdir
		pasea=directory_walker(self.srcdir)
		directorios=pasea.take_directory_listing()
		longitud=len(self.srcdir)
		while 0!=len(directorios):			
			for fichero in directorios[1:]:
				if self.ficheros.has_key(fichero):
					print "Warning: file /"+directorios[0][longitud:]+fichero+" conflicts with /"+self.ficheros[fichero]
				else:
					self.ficheros[fichero]=directorios[0][longitud:]+fichero
			directorios=pasea.take_directory_listing()

	
	def search_file(self,filename):
		"""returns the file's path given a filename, or empty if that filename doesn't exists"""
		if self.ficheros.has_key(filename):
			return self.ficheros[filename]
		else:
			return ""


	def size(self,picture_path):
		"""returns the size of the given picture"""
	
		if self.diccionario.has_key(picture_path):
			pic_params=self.diccionario[picture_path]
			anadir=False
		else:
			identify=subprocess.Popen('identify -format "%w %h x" '+str(picture_path),stdout=subprocess.PIPE,shell=True)
			#std_in,std_out=os.popen2("identify -format \"%w %h x\" "+picture_path)
			pic_params=identify.stdout.readline()
			identify.wait()
			anadir=True
		pos=string.find(pic_params," ",0)
		pos2=string.find(pic_params," ",pos+1)
		if pos!=-1:
			if anadir:
				self.diccionario[picture_path]=pic_params
			return pic_params[:pos],pic_params[pos+1:pos2]
		else:
			return "",""

	
	def test_extension(self,filename,extension_group):
		""" returns True if the file's extension is in the EXTENSION_GROUP list"""
		
		for extens in extension_group:			
			if extens==filename[-len(extens):].lower():
				return True
		return False
	
	
	def test_protocol(self,filename,protocol_group):
		"""returns True if the file start with one element of the PROTOCOL_GROUP list"""
	
		for protocol in protocol_group:
			if (protocol+":")==filename[:1+len(protocol)]:
				return True
		return False

	def trocea(self,file_input,script=False):
		"""self function cuts the file, searching the HTML components
		in base to MODO (0=search < ; 1=search >," or space ; 2=search close " )
		It returns one element each time is called. """
		
		if self.cut_cr: # if the last element had a CR, we send it now
			self.cut_cr=False
			return "\n"
		
		sigue=True
		
		if self.cut_corte>=self.cut_longitud:
			linea=file_input.readline()
			if linea=="":
				return "" # return an end of file
			self.cut_linea=linea			
			self.cut_longitud=len(self.cut_linea)
			self.cut_corte=0
		else:
			linea=self.cut_linea
			
		if self.cut_modo==0:
			corte2=string.find(linea,"<",self.cut_corte)
			if corte2==-1:
				salida=linea[self.cut_corte:]
				self.cut_corte=self.cut_longitud
			else:
				if self.cut_corte!=corte2:
					salida=linea[self.cut_corte:corte2]
					sigue=False
				self.cut_corte=corte2
				self.cut_modo=1				
		if self.cut_modo==1 and sigue:
			corte2=string.find(linea," ",self.cut_corte)
			corte3=string.find(linea,">",self.cut_corte)
			corte4=string.find(linea,"\"",self.cut_corte)
			if corte2==-1 and corte3==-1 and corte4==-1:
				salida=linea[self.cut_corte:]
				self.cut_corte=self.cut_longitud
			else:
				if corte2==-1:
					corte2=self.cut_longitud+1
				if corte3==-1:
					corte3=self.cut_longitud+1
				if corte4==-1:
					corte4=self.cut_longitud+1
				if corte2>corte3:
					corte2=corte3
					self.cut_modo=0
				if corte2>corte4:
					corte2=corte4
					self.cut_modo=2
				corte2=corte2+1
				salida=linea[self.cut_corte:corte2]
				self.cut_corte=corte2
		elif self.cut_modo==2:
			corte2=string.find(linea,'"',self.cut_corte)
			if script:
				corte3=1+string.find(linea,'\\',self.cut_corte)
			else:
				corte3=corte2
			if corte2==-1:					
				salida=linea[self.cut_corte:]
				self.cut_corte=self.cut_longitud
			else:
				salida=linea[self.cut_corte:corte2+1]
				self.cut_corte=corte2+1
				if corte2==corte3: # we test for \", since it isn't an end,
					self.cut_modo=1 # but a " in PHP, or JavaScript (but only if SCRIPT is true)
		if salida=="\n" or salida=="\r\n":
			return "\n"

		if salida[-1]=="\n": # if the element ends in a Carriage Return
			if salida[-2]=="\r":
				salida=salida[:-2] # delete CR and LF
			else:			
				salida=salida[:-1] # delete CR
			self.cut_cr=True # and remember to send it alone next time
		
		return salida
		
	def check_exists_file(self,base,fichero):
	
		if self.test_protocol(fichero,self.protocols):
			return True # if it's a protocol (http, ftp...), it always exists
		
		temporal=base+fichero
		os.path.normpath(temporal)
		return os.path.isfile(temporal)
		
	def write_output(self,file_output,element):
		""" writes the element to the output file and adds a CR if needed """
	
		file_output.write(element)
		
	def procesa_ficheros(self,entrada,dir_base,file_output,profundidad):
		""" Process the ENTRADA file, writting the output in FILE_OUTPUT """

		file_input=open(entrada, 'r')

		self.cut_linea=""
		self.cut_modo=0
		self.cut_corte=0
		self.cut_longitud=0 # init for TROCEA function
		self.cut_cr=False;

		self.fich_salida=file_output

		elemento=self.trocea(file_input)
		while elemento!="":			
			if elemento=="<!-- ":
				self.tag_comment(file_input,dir_base,file_output,profundidad)
			else:
				self.write_output(file_output,elemento)
				if elemento=="<a ":
					self.tag_a(file_input,dir_base,file_output,profundidad)
				elif elemento=="<area ":
					self.tag_a(file_input,dir_base,file_output,profundidad)
				elif elemento=="<img ":
					self.tag_img(file_input,dir_base,file_output,profundidad)
				elif elemento=="<link ":
					self.tag_a(file_input,dir_base,file_output,profundidad)
				elif elemento=="<script>":
					self.tag_script(file_input,dir_base,file_output,profundidad)
				elif elemento=="<?php ":
					self.tag_php(file_input,dir_base,file_output,profundidad)
				elif elemento[0]=="<" and elemento[-1]!=">":
					self.tag_opentag(file_input,dir_base,file_output,profundidad)
			elemento=self.trocea(file_input)
		file_input.close()

	def get_filename(self,file_input,dir_base,profundidad):
		""" Reads a filename in HTMLProc convention and converts it into a full filename """
		
		elemento=self.trocea(file_input)
		if elemento=="":
			print ("Unexpected end of file (filename)")
			return ""

		if elemento[0]=="/":
			fichero="../"*profundidad+elemento[1:-1] # drop out the " at the end
		elif elemento[0]=="*":
			fichero2=self.search_file(elemento[1:-1])
			if fichero2!="":
				fichero="../"*profundidad+fichero2
			else:
				fichero=elemento[1:-1]
		else:
			fichero=elemento[:-1]		
		return fichero,elemento

	def tag_script(self,file_input,dir_base,file_output,profundidad):
		""" Process the script tags """
		
		elemento=self.trocea(file_input,True) # we are reading from a script tag, so we must
		# have in mind that \" is not a "close quotes"
		while elemento!="":
			self.write_output(file_output,elemento)
			if elemento=="</script>":
				return
			elemento=self.trocea(file_input,True)		
		
		print "Unexpected end of file (script)"
		
	def tag_php(self,file_input,dir_base,file_output,profundidad):
		""" Process the php tags """
		
		elemento=self.trocea(file_input,True) # we are reading from a php tag, so we must
		# have in mind that \" is not a "close quotes"
		while elemento!="":
			self.write_output(file_output,elemento)
			if elemento[-2:]=="?>":
				return
			elemento=self.trocea(file_input,True)		
		
		print "Unexpected end of file (php)"
		

	def tag_comment(self,file_input,dir_base,file_output,profundidad):
		""" Process the comment tags, differentiating between true comments and
		INCLUDE commands """

		elemento=self.trocea(file_input)
		if elemento!="":
			if elemento=="include=\"": # if it's an INCLUDE, include it
				acomment=False
				fichero,f_elemento=self.get_filename(file_input,dir_base,profundidad)
				if False==os.path.isfile(dir_base+fichero): # test if the file exists
					print "Can't find file to include "+fichero
				else:
					if self.test_extension(fichero,self.exclude): #test if we must exclude this from being processed
						fichero_temporal=open(dir_base+fichero,"r")
						for linea2 in fichero_temporal:
							file_output.write(linea2) # copy the file "as is"
						fichero_temporal.close()
					else:
						proc2=processor("") # the included file must be processed too
						proc2.procesa_ficheros(dir_base+fichero,dir_base,file_output,profundidad)
			else:
				self.write_output(file_output,"<!-- "+elemento)
				acomment=True

			elemento=self.trocea(file_input)
			while elemento!="":
				if acomment: # copy all only if it's a comment, no if it's an include
					self.write_output(file_output,elemento)
				if elemento[-3:]=="-->":
					return
				elemento=self.trocea(file_input)
		
		print("Unexpected end of file (comment)")

	def tag_a(self,file_input,dir_base,file_output,profundidad):
		""" Process the a, area and link tags """
	
		elemento=self.trocea(file_input)
		while elemento!="":
			self.write_output(file_output,elemento)
			if elemento=="href=\"":
				fichero,f_elemento=self.get_filename(file_input,dir_base,profundidad)
				self.write_output(file_output,fichero+"\"")
				if (fichero[0]!="#") and (False==self.check_exists_file(dir_base,fichero)):
					print "Warning: file "+fichero+" not found."
			if elemento[-1]==">":
				return
			elemento=self.trocea(file_input)
		print("Unexpected end of file (a)")

	def tag_img(self,file_input,dir_base,file_output,profundidad):
		""" Process the img tags """
	
		elemento=self.trocea(file_input)
		while elemento!="":
			self.write_output(file_output,elemento)
			if elemento=="src=\"":
				fichero,f_elemento=self.get_filename(file_input,dir_base,profundidad)
				self.write_output(file_output,fichero+"\"")
				if f_elemento[0]=="/" or f_elemento[0]=="*":
					f_elemento=f_elemento[1:]
				if False==self.check_exists_file(dir_base,fichero):
					print "Warning: file "+fichero+" not found."				
				elif self.test_extension(f_elemento[:-1],self.pictures) and (False==self.test_protocol(f_elemento,self.protocols)) and self.identify_pic: # is a picture?
					temporal=dir_base+fichero
					os.path.normpath(temporal)
					x,y=self.size(temporal)
					if x!="":					
						self.write_output(file_output," width=\""+x+"\" height=\""+y+"\"")					
			if elemento[-1]==">":
				return
			elemento=self.trocea(file_input)
		print("Unexpected end of file (img)")

	def tag_opentag(self,file_input,dir_base,file_output,profundidad):
		""" Process all the other tags """

		elemento=self.trocea(file_input)
		while elemento!="":
			self.write_output(file_output,elemento)
			if elemento[-1]==">":
				return
			elemento=self.trocea(file_input)
		print("Unexpected end of file (opentag)")

	def process_html(self):
	
		source_directory=self.srcdir
		final_directory=self.finaldir

		recorre=directory_walker(source_directory) # we start in source directory
		
		len_source=len(source_directory)

		directorios=recorre.take_directory_listing()
		while 0!=len(directorios):
			dir_extra=(directorios[0])[len_source:]
			profundidad=string.count(dir_extra,"/")			
			if False==os.path.exists(final_directory+dir_extra):
				os.mkdir(final_directory+dir_extra)
			for fichero in directorios[1:]:
				fichero_final=final_directory+dir_extra+fichero
				
				if self.test_extension(fichero,self.extensions):
					hacer=True
					if os.path.exists(fichero_final):
						estado1=os.stat(directorios[0]+fichero)
						estado2=os.stat(final_directory+dir_extra+fichero)
						if estado1.st_mtime<=estado2.st_mtime:
							hacer=False
					if hacer:
						print "Processing file: "+dir_extra+fichero
						salida=open(final_directory+dir_extra+fichero,"w") # output file
						self.procesa_ficheros(directorios[0]+fichero,directorios[0],salida,profundidad)
						salida.close();
					else:
						print "Skipping file: "+dir_extra+fichero
				elif self.test_extension(fichero,self.tolink) or self.test_extension(fichero,self.pictures):
					if self.makecopy:
						copia=True
						if os.path.isfile(fichero_final):
							estado1=os.stat(directorios[0]+fichero)
							estado2=os.stat(fichero_final)
							if estado1.st_mtime<=estado2.st_mtime:
								copia=False
						if copia:
							print "Copying file: "+dir_extra+fichero
							shutil.copyfile(directorios[0]+fichero,final_directory+dir_extra+fichero)
					else:
						if os.path.isfile(fichero_final):
							os.remove(fichero_final)
						print "Linking file: "+fichero_final
						os.symlink(directorios[0]+fichero,fichero_final)

			directorios=recorre.take_directory_listing()
		


if len(sys.argv)>2:
	print ("Usage: htmlproc config_file")
	sys.exit()

if len(sys.argv)==2:
	fichero=sys.argv[1]
else:
	fichero="config"

print "HTMLProc V1.4"

procesador=processor(fichero)
procesador.process_html()

print "Done"
